#include "kwil_generate.h"
#include <inttypes.h>
#include <stdlib.h>
#include <string.h>

static
int kwil_header_generated_file_name(struct kwil_header_t* self, char** out) {
    int len = strlen(self->file_name);
    char* filename = malloc(len + 5);
    strncpy(filename, self->file_name, len-2);
    strcpy(filename + len-2, ".kwil.h");
    *out = filename;
    return len;
}

static
int kwil_typename_to_prefix(const char* restrict type_name, char* out_prefix, size_t out_size) {
    if(strncmp("enum ", type_name, 5) == 0) {
        type_name += 5;
    } else if(strncmp("struct ", type_name, 7) == 0) {
        type_name += 7;
    }

    int len = strlen(type_name);
    if(len > out_size) {
        return 1;
    }

    // generate a function prefix for a type name
    strcpy(out_prefix, type_name);

    // remove _t postfix if relevant
    if(strcmp(out_prefix + len -3, "_t") == 0) {
        *(out_prefix + len - 3) = '\0';
    }
    return strlen(out_prefix);
}

int kwil_header_generate(struct kwil_header_t* self) {
    char file_buffer[255];
    char* filename;
    kwil_header_generated_file_name(self, &filename);

    FILE* file = fopen(filename, "w+");
    setvbuf(file, file_buffer, _IOLBF, 255);

    if(!file) {
        printf("Failed to open %s for writing", filename);
        free(filename);
        return 1;
    }

    fprintf(file, "#include <string.h>\n");

    fprintf(file, "#define KWIL_GEN_IMPL(...)\\\n");

    for(int type_index = 0; type_index < self->types_len; ++type_index) {
        struct kwil_type_t* type = self->types + type_index;
        printf("Generating functions for %s\n", type->type_name);
        switch(type->type_tag) {
            case KWIL_TYPE_STRUCT:
                kwil_struct_generate_json_length(&type->struct_type, file, type->type_name);
                kwil_struct_generate_to_json(&type->struct_type, file, type->type_name);
                break;
            case KWIL_TYPE_ENUM:
                kwil_enum_generate_json_length(&type->enum_type, file, type->type_name);
                kwil_enum_generate_to_json(&type->enum_type, file, type->type_name);
                break;
        }
    }

    fprintf(file, "\n");
    fclose(file);
    free(filename);

    return 0;
}

static
void kwil_field_generate_to_string(struct kwil_field_t* self, FILE* file) {
    char custom_type_prefix[48];
    switch(self->type_tag) {
        case KWIL_FIELD_CHAR:
            if(self->array_length > 0 || self->array_dynamic) {
                fprintf(file, "    if(src->%s != NULL) {\\\n"
                              "        json_len += sprintf(json + json_len, \"\\\"%%s\\\"\", src->%s);\\\n"
                              "    }\\\n",
                              self->name_str, self->name_str);
            }
            break;
        case KWIL_FIELD_INT:
            fprintf(file, "    json_len += sprintf(json + json_len, \"%%d\", src->%s);\\\n", self->name_str);
            break;
        case KWIL_FIELD_FLOAT:
            fprintf(file, "    json_len += sprintf(json + json_len, \"%%f\", src->%s);\\\n", self->name_str);
            break;
        case KWIL_FIELD_UNSIGNED:
            fprintf(file, "    json_len += sprintf(json + json_len, \"%%u\", src->%s);\\\n", self->name_str);
            break;
        case KWIL_FIELD_UNKNOWN:
        case KWIL_FIELD_CUSTOM:
            kwil_typename_to_prefix(self->type_str, custom_type_prefix, 47);
            fprintf(file, "    json_len += %s_to_json(&src->%s, json + json_len);\\\n", custom_type_prefix, self->name_str);
            break;
    }
}

static
void kwil_field_get_serialized_length(struct kwil_field_t* self, FILE* file) {
    char custom_type_prefix[48];
    size_t len = strlen(self->name_str) + 4; // "name": ,

    fprintf(file, "    /* length of %s */\\\n"
                  "    json_capacity += %zu",
                  self->name_str, len);

    switch(self->type_tag) {
        case KWIL_FIELD_CHAR:
            if(self->array_dynamic || self->array_length > 0) {
                fprintf(file, " + strlen(src->%s);\\\n", self->name_str);
            } else {
                fprintf(file, " + 3;\\\n");
            }
            break;
        case KWIL_FIELD_INT:
            fprintf(file, "+ snprintf(NULL, 0, \"%%d\", src->%s);\\\n", self->name_str);
            break;
        case KWIL_FIELD_UNSIGNED:
            fprintf(file, " + snprintf(NULL, 0, \"%%du\", src->%s);\\\n", self->name_str);
            break;
        case KWIL_FIELD_FLOAT:
            fprintf(file, " + snprintf(NULL, 0, \"%%f\", src->%s);\\\n", self->name_str);
            break;
        case KWIL_FIELD_CUSTOM:
        case KWIL_FIELD_UNKNOWN:
            kwil_typename_to_prefix(self->type_str, custom_type_prefix, 47);
            fprintf(file, " + %s_json_length(&src->%s);\\\n", custom_type_prefix, self->name_str);
            break;
    }
}

int kwil_struct_generate_json_length(struct kwil_struct_t* self, FILE* file, const char* type_name) {
    char prefix[48];
    kwil_typename_to_prefix(type_name, prefix, 47);

    fprintf(file, "size_t %s_json_length(struct %s* src) {\\\n"
                  "    size_t json_capacity = 2;\\\n",// allocate at least two for "{}\0"
            prefix, type_name);
    for(int field_index = 0; field_index < self->fields_len; ++field_index) {
        kwil_field_get_serialized_length(self->fields + field_index, file);
    }
    fprintf(file, "    return json_capacity;\\\n"
                  "}\\\n");
    return 0;
}

int kwil_struct_generate_to_json(struct kwil_struct_t* self, FILE* file, const char* type_name) {
    char prefix[48];
    kwil_typename_to_prefix(type_name, prefix, 47);

    fprintf(file, "int %s_to_json(struct %s* src, char* json) {\\\n"
                  , prefix, type_name);

    fprintf(file, "    int json_len = 1;\\\n"
                  "    strcpy(json, \"{\");\\\n");

    for(int field_index = 0; field_index < self->fields_len; ++field_index) {
        kwil_field_generate_to_json(self->fields + field_index, file);
    }

    fprintf(file, "    strcpy(json + json_len - 1, \"}\");\\\n"
                  "    return json_len;\\\n"
                  "}\\\n");
    fflush(file);

    return 0;
}

int kwil_field_generate_to_json(struct kwil_field_t* self, FILE* file) {
    fprintf(file, "    /* field: %s */\\\n", self->name_str);
    fprintf(file, "    json_len += sprintf(json + json_len, \"\\\"%s\\\":\");\\\n",
                  self->name_str);
    kwil_field_generate_to_string(self, file);
    fprintf(file, "    strcpy(json + json_len, \",\");\\\n"
                  "    ++json_len;\\\n");
    return 0;
}

int kwil_enum_generate_json_length(struct kwil_enum_t* self, FILE* file, const char* type_name) {
    char prefix[48];
    kwil_typename_to_prefix(type_name, prefix, 47);
    fprintf(file, "size_t %s_json_length(enum %s* src) {\\\n"
                  "    switch(*src) {\\\n", prefix, type_name);
    for(int option_index = 0; option_index < self->enum_values_len; ++option_index) {
        fprintf(file, "        case %s: return %zu;\\\n",
                self->enum_values[option_index].name, strlen(self->enum_values[option_index].name));
    }
    fprintf(file, "    }\\\n"
                  "    return 2;\\\n"
                  "}\\\n");
    return 0;
}

int kwil_enum_generate_to_json(struct kwil_enum_t* self, FILE* file, const char* type_name) {
    char prefix[48];
    kwil_typename_to_prefix(type_name, prefix, 47);
    fprintf(file, "size_t %s_to_json(enum %s* src, char* out) {\\\n"
                  "    switch(*src) {\\\n", prefix, type_name);

    for(int option_index = 0; option_index < self->enum_values_len; ++option_index) {
        fprintf(file, "        case %s:\\\n"
                      "            return sprintf(out, \"\\\"%s\\\"\");\\\n",
                self->enum_values[option_index].name, self->enum_values[option_index].name);
    }

    fprintf(file, "    }\\\n"
                  "    return sprintf(out, \"\\\"\\\"\");\\\n"
                  "}\\\n");
    return 0;
}

